from zeep.client import CachingClient, Client  # noqa
from zeep.transports import Transport  # noqa
from zeep.plugins import Plugin  # noqa
from zeep.xsd.valueobjects import AnyObject  # noqa

__version__ = '2.1.1'
